/*******************************************************************************
 Graph Project
 Copyright (C) 2006  denton

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************
 File: src/Parser.java
 Version: 0.1.0  Date: 27.07.2006
 Author: denton
 *******************************************************************************/

package Graph;

public class Parser {

    public static final int e_Unknown = 0;
    public static final int e_DivisionByZero = 100;
    public static final int e_LogExpressionLessZero = 101;
    public static final int e_LogBaseEqualsOne = 102;
    public static final int e_SqrtBaseLessZero = 103;
    public static final int e_TgIsNaN = 104;
    public static final int e_CtgIsNaN = 105;
    public static final int e_ArcsinArgMoreOne = 106;
    public static final int e_ArcsinArgLessOne = 107;
    public static final int e_ArccosArgMoreOne = 108;
    public static final int e_ArccosArgLessOne = 109;
    public static final int e_ZeroPowInZero = 110;
    public static final int e_ZeroPowInNegative = 111;
    public static final int e_InvalidExpression = 200;
    

    public static String insert(String subs, String s, int start) {
        String rs = "";
        for (int i=0;i<start;i++) {
            rs = rs+s.charAt(i);
        }
        for (int i=start;i<start+subs.length();i++) {
            rs = rs+subs.charAt(i-start);
        }
        for (int i=start+subs.length();i<s.length()+subs.length();i++) {
            rs = rs+s.charAt(i-subs.length());
        }
        return rs;
    }
    
    public static String delete(String s, int start, int len) {
        String rs = "";
        for (int i=0;i<start;i++) {
            rs = rs+s.charAt(i);
        }
        for (int i=start+len;i<s.length();i++) {
            rs = rs+s.charAt(i);
        }
        return rs;
    }
    
    public static String replacestr(String s, String subs, String ends) {
        int pos = s.indexOf(subs);
        if (pos>=0) {
            s = delete(s,pos,subs.length());
            s = insert(ends,s,pos);
        }
        if (s.indexOf(subs)>=0) s = replacestr(s,subs,ends);
        return s;
    }

    public static int isInString(String s, char c) {
        int x = 0;
        for (int i=0;i<s.length();i++) {if (s.charAt(i)==c) x++;}
        return x;
    }



    public static boolean isCharOfFunction(char c) {
        return ((c=='n')|(c=='g')|(c=='s')|(c=='t'));
    }
    
    public static boolean isNumber(char c) {
        return ((c=='0')|(c=='1')|(c=='2')|(c=='3')|(c=='4')|(c=='5')|(c=='6')|(c=='7')|(c=='8')|(c=='9')|(c=='E'));
    }
    
    public static boolean isSign(char c) {
        return ((c=='+')|(c=='-')|(c=='*')|(c=='/'));
    }

    public static void checkSyntax(String s) throws GraphException {
        if (isInString(s,'(')!=isInString(s,')')) throw (new GraphException(e_InvalidExpression));
        if ((s.indexOf('(')>=0)|(s.indexOf(')')>=0)) {
            int incl = 0;
            for (int i=0;i<s.length();i++) {
                if (s.charAt(i)=='(') incl++;
                if (s.charAt(i)==')') incl--;
                if (incl<0) throw (new GraphException(e_InvalidExpression));
            }
        }
    }
    
    public static String replaceBrackets(String s) {
        do {
            int p1 = s.indexOf("(");
            if (p1>0) {
                if (isCharOfFunction(s.charAt(p1-1))) {
                    int incl = 0, p2 = 0;
                    for (int i=p1;i<=s.length();i++) {
                        if (s.charAt(i)=='(') incl++;
                        if (s.charAt(i)==')') incl--;
                        if (incl==0) {
                             p2=i;
                             break;
                        }
                    }
                    s = delete(s,p1,1);
                    s = insert("[",s,p1);
                    s = delete(s,p2,1);
                    s = insert("]",s,p2);
                }
            } else break;
        } while (true);


        return s;

    }

    public static String replaceConstants(String s) {
            while (s.indexOf("e")>=0) s = replacestr(s,"e",Float.E.toString());
            while (s.indexOf("pi")>=0) s = replacestr(s,"pi",Float.PI.toString());
        return s;
    }

    public static Float val(String s) {
        Float x = Float.ZERO;
        boolean min = false;
        String ts = "";
        for (int i=0;i<s.length();i++) {
            char c = s.charAt(i);
            if (isNumber(c)|(c=='.')) 
                ts = ts+c;
            if (c=='-') min = !min;
        }
        x = Float.parse(ts,10);
        // if (x.toDouble()<0.031) x = Float.ZERO;
        if (min) x = x.Neg();
        return x;
    }
    
    
    private static int sp1(String s){
        int x = -1;
        for (int i=s.length()-1;i>=0;i--) {if (s.charAt(i)=='+') {x = i; break;} }
        return x;
    }    
    
    private static int sp2(String s){
        int x = -1;
        for (int i=s.length()-1;i>0;i--) {
            char c = s.charAt(i-1);
            if ((s.charAt(i)=='-')&(isNumber(c)))
                {x = i; break;}
        }
        return x;
    }    
    
    private static int sp3(String s){
        int x = -1;
        for (int i=s.length()-1;i>=0;i--) {if (s.charAt(i)=='*') {x = i; break;} }
        return x;
    }    
    
    private static int sp4(String s){
        int x = -1;
        for (int i=s.length()-1;i>=0;i--) {if (s.charAt(i)=='/') {x = i; break;} }
        return x;
    }

    private static int spp(String s){
        int x = -1;
        for (int i=s.length()-1;i>=0;i--) {if (s.charAt(i)=='^') {x = i; break;} }
        return x;
    }
    
    private static boolean uminus(String s){
        return (s.indexOf("--")==0);
    }
    
    public static String calc(String s) throws GraphException {
        while ((uminus(s))|(sp1(s)>=0)|(sp2(s)>0)|(sp3(s)>=0)|(sp4(s)>=0)|(s.indexOf('|')>=0)|(s.indexOf('[')>=0)|(s.indexOf('^')>=0)) {
            
            if (uminus(s)) {
                String sx = s.substring(1,s.length());
                Float x = val(calc(sx));
                x = x.Neg();
                s = x.toString();
                continue;
            }

            if (s.indexOf('|')>=0) {
                int p1 = s.indexOf('|');
                String sx = s.substring(p1+1,s.length());
                int p2 = sx.indexOf('|')+p1+1;
                if (p2<p1+1) throw (new GraphException(e_InvalidExpression));
                sx = s.substring(p1+1,p2);
                s = delete(s,p1,p2-p1+1);
                Float x = Float.abs(val(calc(sx)));
                s = insert(x.toString(),s,p1);
                continue;           
            }
            
            if (s.indexOf("arcsin[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("arcsin[");
                int c2 = c1;
                for (int i=c1+6;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+7,c2)));
                if (!Float.abs(x).Great(Float.ONE)) {
                    x = Float.asin(x);
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                } else if (x.Great(Float.ONE)) {
                    throw (new GraphException(e_ArcsinArgMoreOne));
                } else throw (new GraphException(e_ArcsinArgLessOne));
                continue;
            }           

            if (s.indexOf("arccos[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("arccos[");
                int c2 = c1;
                for (int i=c1+6;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+7,c2)));
                if (!Float.abs(x).Great(Float.ONE)) {
                    x = Float.acos(x);
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                } else if (x.Great(Float.ONE)) {
                    throw (new GraphException(e_ArccosArgMoreOne));
                } else throw (new GraphException(e_ArccosArgLessOne));
                continue;
            }           

            if (s.indexOf("arctg[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("arctg[");
                int c2 = c1;
                for (int i=c1+5;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+6,c2)));
                x = Float.atan(x);
                s = delete(s,c1,c2-c1+1);
                s = insert(x.toString(),s,c1);
                continue;
            }           

            if (s.indexOf("arcctg[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("arcctg[");
                int c2 = c1;
                for (int i=c1+6;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+7,c2)));
                s = delete(s,c1,c2-c1+1);
                if (!x.Equal(Float.ZERO)) {
                    x = Float.ONE.Div(x);
                    x = Float.atan(x);
                    while (x.Less(0)) x = x.Add(Float.PI);
                    s = insert(x.toString(),s,c1);
                } else s = insert(Float.ZERO.toString(),s,c1);
                continue;
            }           

            if (s.indexOf("sin[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("sin[");
                int c2 = c1;
                for (int i=c1+3;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+4,c2)));
                x = Float.sin(x);
                s = delete(s,c1,c2-c1+1);
                s = insert(x.toString(),s,c1);
                continue;           
            }           
            
            if (s.indexOf("cos[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("cos[");
                int c2 = c1;
                for (int i=c1+3;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+4,c2)));
                x = Float.cos(x);
                s = delete(s,c1,c2-c1+1);
                s = insert(x.toString(),s,c1);
                continue;           
            }           

            if (s.indexOf("ctg[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("ctg[");
                int c2 = c1;
                for (int i=c1+3;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                String sx = s.substring(c1+4,c2);
                Float tant = val(calc(sx));
                Float cost = tant;
                tant = Float.tan(tant);
                cost = Float.cos(cost);

                if (tant.Equal(Float.ZERO)) {
                    throw (new GraphException(e_CtgIsNaN));
                } else if (cost.Equal(Float.ZERO)) {
                    s = delete(s,c1,c2-c1+1);
                    s = insert(Float.ZERO.toString(),s,c1);
                } else {
                    Float x = Float.ONE;
                    x = x.Div(tant);
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                }
                continue;
            }           
            
            if (s.indexOf("tg[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("tg[");
                int c2 = c1;
                for (int i=c1+2;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                String sx = s.substring(c1+3,c2);
                Float sint = val(calc(sx));
                Float cost = sint;
                sint = Float.sin(sint);
                cost = Float.cos(cost);
                if (cost.Equal(Float.ZERO)) {
                    throw (new GraphException(e_TgIsNaN));
                } else if (sint.Equal(Float.ZERO)) {
                    s = delete(s,c1,c2-c1+1);
                    s = insert(Float.ZERO.toString(),s,c1);
                } else {
                    Float x = Float.tan(val(calc(sx)));
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                }
                continue;
            }           
            
            
            if (s.indexOf("sqrt[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("sqrt[");
                int c2 = c1;
                for (int i=c1+4;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float vc = val(calc(s.substring(c1+5,c2)));
                if (!vc.Less(Float.ZERO)) {
                    Float x = Float.sqrt(vc);
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                } else throw (new GraphException(e_SqrtBaseLessZero));
                continue;
            }           

            if (s.indexOf("log[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("log[");
                int c2 = c1, w = c1;
                for (int i=c1+3;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if ((s.charAt(i)==',')&(incl==1)) w=i;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x1 = val(calc(s.substring(c1+4,w)));
                Float x2 = val(calc(s.substring(w+1,c2)));
                if ((x2.Great(Float.ZERO))&(x1.Great(Float.ZERO))&(!x1.Equal(Float.ONE))) {
                    Float x = Float.log(x2);
                    x = x.Div(Float.log(x1));
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                } else if (x1.Equal(Float.ONE)) {
                    throw (new GraphException(e_LogBaseEqualsOne));
                } else throw (new GraphException(e_LogExpressionLessZero));
                continue;           
            }           
            
            if (s.indexOf("lg[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("lg[");
                int c2 = c1;
                for (int i=c1+2;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+3,c2)));
                if (x.Great(Float.ZERO)) {
                    Float fx = new Float(10L);
                    x = Float.log(x);
                    x = x.Div(Float.log(fx));
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                } else throw (new GraphException(e_LogExpressionLessZero));
                continue;           
            }           

            if (s.indexOf("ln[")>=0) {
                int incl = 0;
                int c1 = s.indexOf("ln[");
                int c2 = c1;
                for (int i=c1+2;i<s.length();i++) {
                    if (s.charAt(i)=='[') incl++;
                    if (s.charAt(i)==']') incl--;
                    if (incl==0) {
                        c2 = i;
                        break;
                    }
                }
                Float x = val(calc(s.substring(c1+3,c2)));
                if (x.Great(Float.ZERO)) {
                    x = Float.log(x);
                    s = delete(s,c1,c2-c1+1);
                    s = insert(x.toString(),s,c1);
                } else throw (new GraphException(e_LogExpressionLessZero));
                continue;           
            }           


            
            int c1 = sp1(s), c2 = sp2(s), c3 = sp3(s), c4 = sp4(s);
            if (c1>=0) {
                Float x = val(calc(s.substring(0,c1)));
                x = x.Add(val(calc(s.substring(c1+1,s.length()))));
                s = x.toString();
                continue;
            }
            if (c2>0) {
                Float x = val(calc(s.substring(0,c2)));
                x = x.Sub(val(calc(s.substring(c2+1,s.length()))));
                s = x.toString();
                continue;
            }        
            if (c3>=0) {
                Float x = val(calc(s.substring(0,c3)));
                x = x.Mul(val(calc(s.substring(c3+1,s.length()))));
                s = x.toString();
                continue;
            }        
            if (c4>=0) {
                Float vc = val(calc(s.substring(c4+1,s.length())));
                if (!vc.Equal(Float.ZERO)) {
                    Float x = val(calc(s.substring(0,c4)));
                    x = x.Div(vc);
                    s = x.toString();
                } else throw (new GraphException(e_DivisionByZero));
                continue;
            }        
        
        }
        if (s.indexOf("NaN")>=0) throw (new GraphException());
        return s;
    }
    
    public static String R(String s) throws GraphException {
        while (s.indexOf('^')>=0) {
            int pp = s.indexOf('^');
            int c1 = 0, c2 = s.length()-1, incl = 0;
            for (int i=pp-1;i>0;i--) {
                if (s.charAt(i)=='(') incl++;
                if (s.charAt(i)==')') incl--;
                if (incl==0) {
                    c1 = i;
                    break;
                }
            }
            if (c1==pp-1) {
                int j = 0;
                for (int i=pp-1;i>=0;i--) {
                    if (isSign(s.charAt(i))) {
                        if ((s.charAt(i)!='-')|((s.charAt(i)=='-')&(i==0))|((s.charAt(i)=='-')&(isNumber(s.charAt(i-1))))) {
                            c1 = i+1;
                            break;
                        }
                    }
                    j = i;
                }
                if (j==0) c1 = 0;
            }
            incl = 0;
            for (int i=pp+1;i<s.length();i++) {
                if (s.charAt(i)=='(') incl++;
                if (s.charAt(i)==')') incl--;
                if (incl==0) {
                    c2 = i;
                    break;
                }
            }
            if (c2==pp+1) {
                int j = 0;
                for (int i=pp+1;i<s.length();i++) {
                    if (isSign(s.charAt(i))) {
                        if ((s.charAt(i)!='-')|((s.charAt(i)=='-')&(isNumber(s.charAt(i-1))))) {
                            c2 = i-1;
                            break;
                        }
                    }
                    j = i;
                }
                if (j==s.length()-1) c2 = s.length()-1;
            }
            
//            Integer C1 = new Integer(c1);
//            Integer PP = new Integer(pp);
//            Integer C2 = new Integer(c2);
//            javax.microedition.lcdui.Alert a = new javax.microedition.lcdui.Alert("Debug stopped","c1="+C1.toString()+"; pp="+PP.toString()+"; c2="+C2.toString(),null,null);
//            Graph.display.setCurrent(a);
            
            String ss = s.substring(c1,c2+1);
            s = delete(s,c1,c2-c1+1);
            Float n1 = val(R(ss.substring(0,pp-c1)));
            Float n2 = val(R(ss.substring(pp-c1+1,c2-c1+1)));
//            if ((n2.toDouble()!=n2.toLong())&(n1.Less(Float.ZERO))) n1 = Float.abs(n1);
            if ((n2.m_E!=0)&(n1.Less(Float.ZERO))) n1 = Float.abs(n1);
            if ((n1.Equal(n2))&(n1.Equal(Float.ZERO))) throw (new GraphException(e_ZeroPowInZero));
            if ((n1.Equal(Float.ZERO))&(n2.Less(Float.ZERO))) throw (new GraphException(e_ZeroPowInNegative));
            n1 = n1.pow(n1,n2);
            s = insert(n1.toString(),s,c1);
        }
        
        while (s.indexOf('(')>=0) {
            int incl = 0;
            int c1 = s.indexOf('('), c2 = c1;
            for (int i=c1;i<s.length();i++) {
                if (s.charAt(i)=='(') incl++;
                if (s.charAt(i)==')') incl--;
                if (incl==0) {
                    c2 = i;
                    break;
                }
                if (incl<0) throw (new GraphException(e_InvalidExpression));
            }
            String rs = s.substring(c1+1,c2);
            s = delete(s,c1,c2-c1+1);
            s = insert(R(rs),s,c1);
        }
        try {
            return calc(s);
        } catch (GraphException e) {throw e;}
    }
    
    public static Float calculate(String s) throws GraphException {
        try {
            return Float.parse(R(s),10);
        } catch (GraphException e) {throw e;}
    }

}
